from datetime import datetime, timezone
from unittest.mock import patch

from django.db import IntegrityError
from django.urls import reverse
from rest_framework import status

from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import MetricKey, MetricMode, MetricValue, Model, ModelVersion
from arkindex.users.models import Right, User


class TestMetricsAPI(FixtureAPITestCase):
    """
    Test metrics
    """
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.model = Model.objects.create(name="A Model")
        cls.model_version = ModelVersion.objects.create(
            model=cls.model,
            hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            archive_hash="5eca9bd3eb07c006cd43ae48dfde7fd3",
            size=8,
        )
        cls.user = User.objects.get(email="user@user.fr")
        Right.objects.create(user=cls.user, content_object=cls.model, level=100)
        cls.metric_key = MetricKey.objects.create(name="a test metric", mode=MetricMode.Point, model_version=cls.model_version)

    def test_create_metric_value_requires_logged_in(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "a test metric",
                    "value": 1.2
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_create_metric_value_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "a test metric",
                    "value": 1.2
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_create_metric_value_model_version_not_found(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                    "name": "a test metric",
                    "value": 1.2
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"model_version_id": ['Invalid pk "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa" - object does not exist.']})

    def test_create_metric_value_requires_contributor(self):
        self.user.rights.all().delete()
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "a test metric",
                    "value": 1.2
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have contributor access to this model."})

    def test_create_metric_value_archived(self):
        self.model.archived = datetime.now(timezone.utc)
        self.model.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "a test metric",
                    "value": 1.2
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"model_version_id": ["This ModelVersion is part of an archived model."]})

    @patch("django.utils.timezone.now")
    def test_create_metric_value_existing_metric_key(self, datetime_mock):
        """
        It is possible to create a metric value for an existing metric key.
        It does not create a new metric key.
        """
        datetime_mock.return_value = datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc)
        self.client.force_login(self.user)
        self.assertEqual(len(MetricKey.objects.filter(model_version=self.model_version)), 1)
        self.assertEqual(len(MetricValue.objects.filter(metric=self.metric_key)), 0)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "a test metric",
                    "value": 1.2
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertDictEqual(response.json(), {"created": "2046-01-01T12:34:56Z", "step": None, "value": 1.2})
        self.assertEqual(len(MetricKey.objects.filter(model_version=self.model_version)), 1)
        self.assertEqual(len(MetricValue.objects.filter(metric=self.metric_key)), 1)
        created_value = MetricValue.objects.get(metric=self.metric_key)
        self.assertEqual(created_value.value, 1.2)
        self.assertEqual(created_value.step, None)
        self.assertEqual(created_value.created, datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc))

    @patch("django.utils.timezone.now")
    def test_create_metric_value_no_metric_key(self, datetime_mock):
        """
        Creating a metric value creates the metric key if it doesn't already exist.
        """
        datetime_mock.return_value = datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc)
        self.client.force_login(self.user)
        with self.assertRaises(MetricKey.DoesNotExist):
            MetricKey.objects.get(name="my_metric")
        with self.assertNumQueries(11):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "my_metric",
                    "value": 1.2
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertDictEqual(response.json(), {"created": "2046-01-01T12:34:56Z", "step": None, "value": 1.2})
        # Check the metric key
        metric_key = (MetricKey.objects.get(name="my_metric"))
        self.assertEqual(metric_key.name, "my_metric")
        self.assertEqual(metric_key.model_version, self.model_version)
        self.assertEqual(metric_key.mode, MetricMode.Series)
        # Check the metric value
        created_value = MetricValue.objects.get(metric=metric_key)
        self.assertEqual(created_value.value, 1.2)
        self.assertEqual(created_value.step, None)
        self.assertEqual(created_value.created, datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc))

    @patch("django.utils.timezone.now")
    def test_create_metric_value_new_metric_key_point_mode(self, datetime_mock):
        datetime_mock.return_value = datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc)
        self.client.force_login(self.user)
        with self.assertRaises(MetricKey.DoesNotExist):
            MetricKey.objects.get(name="my_metric")
        with self.assertNumQueries(11):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "my_metric",
                    "value": 1.2,
                    "mode": "point"
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertDictEqual(response.json(), {"created": "2046-01-01T12:34:56Z", "step": None, "value": 1.2})
        # Check the metric key
        metric_key = (MetricKey.objects.get(name="my_metric"))
        self.assertEqual(metric_key.name, "my_metric")
        self.assertEqual(metric_key.model_version, self.model_version)
        self.assertEqual(metric_key.mode, MetricMode.Point)
        # Check the metric value
        created_value = MetricValue.objects.get(metric=metric_key)
        self.assertEqual(created_value.value, 1.2)
        self.assertEqual(created_value.step, None)
        self.assertEqual(created_value.created, datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc))

    def test_create_metric_value_new_metric_key_point_step(self):
        """
        Cannot create a metric value of type point using the step property
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(10):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "my_metric",
                    "value": 1.2,
                    "mode": "point",
                    "step": 1
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"step": ["The step property cannot be set for Point type metrics."]})

    def test_create_metric_value_point_step(self):
        """
        Cannot create a metric value using the step property for a metric key of type point
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "a test metric",
                    "value": 1.2,
                    "step": 1
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"step": ["The step property cannot be set for Point type metrics."]})

    def test_create_metric_value_wrong_mode(self):
        """
        Cannot create a metric value for an existing metric key with a different mode
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "a test metric",
                    "value": 1.2,
                    "mode": "series"
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"mode": ["Metric Key a test metric already exists and is of mode Point."]})

    def test_create_metric_value_only_one_point(self):
        """
        A point mode metric key can have only one value
        """
        MetricValue.objects.create(metric=self.metric_key, value=2.3)
        self.assertEqual(MetricKey.objects.get(name="a test metric").values.count(), 1)
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "a test metric",
                    "value": 1.2
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"__all__": ["Metric Key a test metric is of mode Point and already has a value."]})
        self.assertEqual(MetricKey.objects.get(name="a test metric").values.count(), 1)

    def test_create_metric_value_metric_key_name_version_unique(self):
        """
        There cannot be two metric keys with the same name on one model version
        """
        with self.assertRaises(IntegrityError):
            MetricKey.objects.create(name="a test metric", model_version=self.model_version, mode=MetricMode.Series)

    @patch("django.utils.timezone.now")
    def test_create_metric_value_series(self, datetime_mock):
        """
        It is possible to create multiple metric values (different steps) for series mode metric keys
        """
        datetime_mock.return_value = datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc)
        metric_key = MetricKey.objects.create(name="a series metric", mode=MetricMode.Series, model_version=self.model_version)
        MetricValue.objects.create(metric=metric_key, value=3, step=1)
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "a series metric",
                    "value": 1.2,
                    "step": 2
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertDictEqual(response.json(), {"created": "2046-01-01T12:34:56Z", "step": 2, "value": 1.2})

        metric_key_values = MetricValue.objects.filter(metric=metric_key).order_by("step")
        self.assertEqual(len(metric_key_values), 2)
        self.assertEqual(metric_key_values[0].value, 3)
        self.assertEqual(metric_key_values[0].step, 1)
        self.assertEqual(metric_key_values[1].value, 1.2)
        self.assertEqual(metric_key_values[1].step, 2)

    def test_create_metric_value_unique_name_step(self):
        """
        There can only be one metric value per step (series type metrics)
        """
        metric_key = MetricKey.objects.create(name="a series metric", mode=MetricMode.Series, model_version=self.model_version)
        MetricValue.objects.create(metric=metric_key, value=3, step=1)
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "a series metric",
                    "value": 1.2,
                    "step": 1
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"step": ["This metric has already been set for this step."]})

    def test_create_metric_value_series_existing_step(self):
        """
        If a metric key already has values, if they have step set then all new metric values must have step set as well.
        """
        series_metric = MetricKey.objects.create(name="another metric", model_version=self.model_version, mode=MetricMode.Series)
        MetricValue.objects.create(metric=series_metric, step=1, value=0.2)
        self.assertEqual(len(MetricValue.objects.filter(metric=series_metric)), 1)

        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "another metric",
                    "value": 1.2,
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"step": ["Cannot add a metric value without a step for metric key another metric because it already has metric values with steps set."]})
        self.assertEqual(len(MetricValue.objects.filter(metric=series_metric)), 1)

    def test_create_metric_value_series_existing_no_step(self):
        """
        If a metric key already has values, if they do not have step set then new metric values cannot have step set.
        """
        series_metric = MetricKey.objects.create(name="another metric", model_version=self.model_version, mode=MetricMode.Series)
        MetricValue.objects.create(metric=series_metric, value=0.2)
        self.assertEqual(len(MetricValue.objects.filter(metric=series_metric)), 1)

        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:metric-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "name": "another metric",
                    "value": 1.2,
                    "step": 1
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"step": ["Cannot add a metric value with step set for metric key another metric because it already has metric values with no step set."]})
        self.assertEqual(len(MetricValue.objects.filter(metric=series_metric)), 1)

    def test_bulk_create_metric_value_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "metrics": [
                        {
                            "name": "a test metric",
                            "value": 12
                        },
                        {
                            "name": "another metric",
                            "value": 3,
                            "mode": "point"
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_bulk_create_metric_value_model_version_not_found(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                    "metrics": [
                        {
                            "name": "a test metric",
                            "value": 12
                        },
                        {
                            "name": "another metric",
                            "value": 3,
                            "mode": "point"
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"model_version_id": ['Invalid pk "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa" - object does not exist.']})

    def test_bulk_create_metric_value_requires_contributor(self):
        self.user.rights.all().delete()
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "metrics": [
                        {
                            "name": "a test metric",
                            "value": 12
                        },
                        {
                            "name": "another metric",
                            "value": 3,
                            "mode": "point"
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have contributor access to this model."})

    def test_bulk_create_metric_value_archived(self):
        self.model.archived = datetime.now(timezone.utc)
        self.model.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "metrics": [
                        {
                            "name": "a test metric",
                            "value": 12
                        },
                        {
                            "name": "another metric",
                            "value": 3,
                            "mode": "point"
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"model_version_id": ["This ModelVersion is part of an archived model."]})

    def test_bulk_create_metric_value_duplicate(self):
        """
        Cannot send two metrics values for the same metric key / name
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "metrics": [
                        {
                            "name": "some metric",
                            "value": 12
                        },
                        {
                            "name": "some metric",
                            "value": 3,
                            "mode": "point"
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"metrics": {"1": {"__all__": ["Cannot send multiple values at once for the same name (some metric)."]}}})

    @patch("django.utils.timezone.now")
    def test_bulk_create_metric_value_existing_metric_key(self, datetime_mock):
        """
        It is possible to bulk create metric values for existing metric keys.
        It does not create new metric keys.
        """
        datetime_mock.return_value = datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc)
        test_metric = MetricKey.objects.create(name="metric numero duo", model_version=self.model_version, mode=MetricMode.Series)
        self.client.force_login(self.user)
        self.assertEqual(len(MetricKey.objects.filter(model_version=self.model_version)), 2)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "metrics": [
                        {
                            "name": "a test metric",
                            "value": 12
                        },
                        {
                            "name": "metric numero duo",
                            "value": 3,
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertDictEqual(response.json(), {"metrics": [{"value": 12.0}, {"value": 3.0}]})

        self.assertEqual(len(MetricKey.objects.filter(model_version=self.model_version)), 2)
        # Check metric values
        metric_values = MetricValue.objects.all().order_by("value")
        self.assertEqual(metric_values[0].value, 3)
        self.assertEqual(metric_values[0].created, datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc))
        self.assertEqual(metric_values[0].metric, test_metric)
        self.assertEqual(metric_values[0].step, None)
        self.assertEqual(metric_values[1].value, 12)
        self.assertEqual(metric_values[1].created, datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc))
        self.assertEqual(metric_values[1].metric, self.metric_key)
        self.assertEqual(metric_values[1].step, None)

    @patch("django.utils.timezone.now")
    def test_bulk_create_metric_value_no_metric_key(self, datetime_mock):
        """
        Creating metric values in bulk creates the metric keys if they don't already exist.
        """
        datetime_mock.return_value = datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc)
        self.client.force_login(self.user)
        with self.assertRaises(MetricKey.DoesNotExist):
            MetricKey.objects.get(name="some metric")
        with self.assertRaises(MetricKey.DoesNotExist):
            MetricKey.objects.get(name="another metric")
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "metrics": [
                        {
                            "name": "some metric",
                            "value": 12
                        },
                        {
                            "name": "another metric",
                            "value": 3,
                            "mode": "point"
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertDictEqual(response.json(), {"metrics": [{"value": 12.0}, {"value": 3.0}]})

        # Check the metric keys
        metric_key_1 = (MetricKey.objects.get(name="some metric"))
        self.assertEqual(metric_key_1.name, "some metric")
        self.assertEqual(metric_key_1.model_version, self.model_version)
        self.assertEqual(metric_key_1.mode, MetricMode.Series)
        metric_key_2 = (MetricKey.objects.get(name="another metric"))
        self.assertEqual(metric_key_2.name, "another metric")
        self.assertEqual(metric_key_2.model_version, self.model_version)
        self.assertEqual(metric_key_2.mode, MetricMode.Point)

        # Check the metric values
        metric_values = MetricValue.objects.all().order_by("value")
        self.assertEqual(metric_values[0].value, 3)
        self.assertEqual(metric_values[0].created, datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc))
        self.assertEqual(metric_values[0].metric, metric_key_2)
        self.assertEqual(metric_values[0].step, None)
        self.assertEqual(metric_values[1].value, 12)
        self.assertEqual(metric_values[1].created, datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc))
        self.assertEqual(metric_values[1].metric, metric_key_1)
        self.assertEqual(metric_values[1].step, None)

    @patch("django.utils.timezone.now")
    def test_bulk_create_metric_value_step_series(self, datetime_mock):
        """
        It is possible to bulk create metrics values of mode series for a given step value.
        """
        test_metric_1 = MetricKey.objects.create(name="metric numero duo", model_version=self.model_version, mode=MetricMode.Series)
        test_metric_2 = MetricKey.objects.create(name="metric number three", model_version=self.model_version, mode=MetricMode.Series)
        datetime_mock.return_value = datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc)
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "step": 3,
                    "metrics": [
                        {
                            "name": "metric number three",
                            "value": 12
                        },
                        {
                            "name": "metric numero duo",
                            "value": 3,
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertDictEqual(response.json(), {"step": 3, "metrics": [{"value": 12.0}, {"value": 3.0}]})

        # Check the metric values
        metric_values = MetricValue.objects.all().order_by("value")
        self.assertEqual(metric_values[0].value, 3)
        self.assertEqual(metric_values[0].created, datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc))
        self.assertEqual(metric_values[0].metric, test_metric_1)
        self.assertEqual(metric_values[0].step, 3)
        self.assertEqual(metric_values[1].value, 12)
        self.assertEqual(metric_values[1].created, datetime(2046, 1, 1, 12, 34, 56, tzinfo=timezone.utc))
        self.assertEqual(metric_values[1].metric, test_metric_2)
        self.assertEqual(metric_values[1].step, 3)

    def test_bulk_create_metric_value_step_point(self):
        """
        It is not possible to bulk create point mode metric values with step set.
        """
        MetricKey.objects.create(name="metric numero duo", model_version=self.model_version, mode=MetricMode.Point)
        MetricKey.objects.create(name="another metric", model_version=self.model_version, mode=MetricMode.Series)
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "step": 3,
                    "metrics": [
                        {
                            "name": "a test metric",
                            "value": 12
                        },
                        {
                            "name": "metric numero duo",
                            "value": 3,
                        },
                        {
                            "name": "another metric",
                            "value": 1.89
                        },
                        {
                            "name": "loss",
                            "value": 0.72,
                            "mode": "point"
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"metrics": {
            "0": {"step": ["The step property cannot be set for Point type metrics."]},
            "1": {"step": ["The step property cannot be set for Point type metrics."]},
            "3": {"step": ["The step property cannot be set for Point type metrics."]}
        }})

    def test_bulk_create_metric_value_only_one_point(self):
        """
        A point mode metric key can have only one value (bulk edition)
        """
        MetricValue.objects.create(metric=self.metric_key, value=2.3)
        self.assertEqual(MetricKey.objects.get(name="a test metric").values.count(), 1)
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "metrics": [
                        {
                            "name": "a test metric",
                            "value": 12
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "metrics": {
                "0": {"__all__": ["Metric Key a test metric is of mode Point and already has a value."]}
            }
        })
        self.assertEqual(MetricKey.objects.get(name="a test metric").values.count(), 1)

    def test_bulk_create_metric_value_wrong_mode(self):
        """
        Cannot create a metric value for an existing metric key with a different mode (bulk edition)
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "metrics": [
                        {
                            "name": "a test metric",
                            "value": 0.12,
                            "mode": "series"
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"metrics": {"0": {"mode": ["Metric Key a test metric already exists and is of mode Point."]}}})

    def test_bulk_create_metric_value_series_one_value_per_step(self):
        """
        Cannot create a metric value for a given metric key and step if one already exists (bulk edition)
        """
        series_metric = MetricKey.objects.create(name="another metric", model_version=self.model_version, mode=MetricMode.Series)
        MetricValue.objects.create(metric=series_metric, step=1, value=0.2)
        self.assertEqual(len(MetricValue.objects.filter(metric=series_metric)), 1)
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "step": 1,
                    "metrics": [
                        {
                            "name": "another metric",
                            "value": 0.12
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"metrics": {"0": {"step": ["This metric has already been set for this step."]}}})
        self.assertEqual(len(MetricValue.objects.filter(metric=series_metric)), 1)

    def test_bulk_create_metric_value_mutiple_errors(self):
        """
        Check error display for metric items in bulk metric value creation.
        """
        metric_duo = MetricKey.objects.create(name="metric numero duo", model_version=self.model_version, mode=MetricMode.Point)
        MetricKey.objects.create(name="another metric", model_version=self.model_version, mode=MetricMode.Series)
        MetricValue.objects.create(metric=metric_duo, value=56)
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "step": 3,
                    "metrics": [
                        {
                            "name": "a test metric",
                            "value": 12
                        },
                        {
                            "name": "metric numero duo",
                            "value": 3,
                        },
                        {
                            "name": "another metric",
                            "value": 1.89
                        },
                        {
                            "name": "loss",
                            "value": 0.72,
                            "mode": "point"
                        },
                        {
                            "name": "metric numero duo",
                            "value": 9
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "metrics": {
                "0": {"step": ["The step property cannot be set for Point type metrics."]},
                "1": {
                    "step": ["The step property cannot be set for Point type metrics."],
                    "__all__": ["Metric Key metric numero duo is of mode Point and already has a value."]
                },
                "3": {"step": ["The step property cannot be set for Point type metrics."]},
                "4": {
                    "__all__": [
                        "Cannot send multiple values at once for the same name (metric numero duo).",
                        "Metric Key metric numero duo is of mode Point and already has a value."
                    ],
                    "step": ["The step property cannot be set for Point type metrics."]
                }
            }
        })

    def test_bulk_create_metric_value_series_existing_step(self):
        """
        If a metric key already has values, if they have step set then all new metric values must have step set as well (bulk edition).
        """
        series_metric = MetricKey.objects.create(name="another metric", model_version=self.model_version, mode=MetricMode.Series)
        MetricValue.objects.create(metric=series_metric, step=1, value=0.2)
        self.assertEqual(len(MetricValue.objects.filter(metric=series_metric)), 1)

        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "metrics": [
                        {
                            "name": "another metric",
                            "value": 0.12
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"metrics": {"0": {"step": ["Cannot add a metric value without a step for metric key another metric because it already has metric values with steps set."]}}})
        self.assertEqual(len(MetricValue.objects.filter(metric=series_metric)), 1)

    def test_bulk_create_metric_value_series_existing_no_step(self):
        """
        If a metric key already has values, if they do not have step set then new metric values cannot have step set (bulk edition).
        """
        series_metric = MetricKey.objects.create(name="another metric", model_version=self.model_version, mode=MetricMode.Series)
        MetricValue.objects.create(metric=series_metric, value=0.2)
        self.assertEqual(len(MetricValue.objects.filter(metric=series_metric)), 1)

        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:metrics-create"),
                data={
                    "model_version_id": str(self.model_version.id),
                    "step": 2,
                    "metrics": [
                        {
                            "name": "another metric",
                            "value": 0.12
                        }
                    ]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"metrics": {"0": {"step": ["Cannot add a metric value with step set for metric key another metric because it already has metric values with no step set."]}}})
        self.assertEqual(len(MetricValue.objects.filter(metric=series_metric)), 1)
