import logging
import urllib.parse
from uuid import UUID

from django.conf import settings
from django.contrib.auth import login, logout
from django.contrib.auth.forms import PasswordResetForm
from django.contrib.auth.tokens import default_token_generator
from django.contrib.contenttypes.models import ContentType
from django.core.mail import send_mail
from django.db.models import Count, Prefetch, Q
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.http import urlsafe_base64_encode
from django_rq.queues import get_queue
from django_rq.settings import QUEUES
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter, extend_schema, extend_schema_view
from rest_framework import serializers, status
from rest_framework.exceptions import AuthenticationFailed, NotFound, PermissionDenied, ValidationError
from rest_framework.generics import (
    CreateAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveDestroyAPIView,
    RetrieveUpdateDestroyAPIView,
)
from rest_framework.permissions import SAFE_METHODS
from rest_framework.response import Response
from rest_framework.views import APIView
from rq.job import JobStatus

from arkindex.documents.models import Corpus
from arkindex.process.models import Worker
from arkindex.project.mixins import ACLMixin, WorkerACLMixin
from arkindex.project.permissions import IsAuthenticatedOrReadOnly, IsVerified
from arkindex.users.models import Group, Right, Role, Scope, User, UserScope
from arkindex.users.serializers import (
    EmailLoginSerializer,
    GenericMembershipSerializer,
    GroupSerializer,
    JobSerializer,
    MemberGroupSerializer,
    MembershipCreateSerializer,
    MembershipSerializer,
    NewUserSerializer,
    PasswordResetConfirmSerializer,
    PasswordResetSerializer,
    UserSerializer,
)
from arkindex.users.utils import RightContent, get_max_level

logger = logging.getLogger(__name__)


@extend_schema(tags=["users"])
@extend_schema_view(
    get=extend_schema(description="Retrieve information about the authenticated user"),
    patch=extend_schema(description="Update a user's password. This action is not allowed to users without confirmed e-mails."),
    put=extend_schema(description="Update a user's password. This action is not allowed to users without confirmed e-mails."),
    delete=extend_schema(operation_id="Logout", description="Log out from the API")
)
class UserRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Retrieve information about the authenticated user
    """
    permission_classes = (IsAuthenticatedOrReadOnly, )
    serializer_class = UserSerializer

    def get_object(self):
        return self.request.user

    def retrieve(self, *args, **kwargs):
        if not self.request.user.is_authenticated:
            # Include features on GET to let the frontend still know the instance features when not logged in
            return Response(
                {
                    "detail": PermissionDenied.default_detail,
                    "features": settings.ARKINDEX_FEATURES,
                },
                status=status.HTTP_403_FORBIDDEN,
            )

        return super().retrieve(*args, **kwargs)

    def perform_destroy(self, instance):
        logout(self.request)


@extend_schema_view(post=extend_schema(
    operation_id="Register",
    responses={201: UserSerializer},
    tags=["users"],
))
class UserCreate(CreateAPIView):
    """
    Register as a new user.
    """
    serializer_class = NewUserSerializer

    def create(self, request, *args, **kwargs):
        if not settings.ARKINDEX_FEATURES["signup"]:
            raise ValidationError(["Registration has been disabled on this instance."])

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        if User.objects.filter(email=serializer.validated_data["email"]).exists():
            raise ValidationError({"email": ["An user with this email address already exists."]})

        user = serializer.save()
        login(self.request, user)

        activation_url = "{}?{}".format(
            self.request.build_absolute_uri(reverse("api:user-token")),
            urllib.parse.urlencode({
                "email": user.email,
                "token": default_token_generator.make_token(user),
            }),
        )
        sent = send_mail(
            # Subject cannot have any newlines
            subject="".join(render_to_string("registration/verification_email_subject.txt").splitlines()),
            message=render_to_string(
                "registration/verification_email.html",
                context={
                    "url": activation_url,
                },
                request=self.request,
            ),
            from_email=None,
            recipient_list=[user.email],
            fail_silently=True,
        )
        if sent == 0:
            logger.error("Failed to send registration email to {}".format(user.email))

        return Response(UserSerializer(user).data, status=status.HTTP_201_CREATED)


@extend_schema_view(post=extend_schema(
    operation_id="Login",
    responses={201: UserSerializer},
    tags=["users"]
))
class UserEmailLogin(CreateAPIView):
    """
    Login using an email and a password
    """
    serializer_class = EmailLoginSerializer

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        try:
            user = User.objects.get_by_natural_key(serializer.validated_data["email"])
        except User.DoesNotExist:
            raise AuthenticationFailed

        if not user.has_usable_password() or not user.check_password(serializer.validated_data["password"]):
            raise AuthenticationFailed

        login(self.request, user)

        headers = self.get_success_headers(serializer.data)
        return Response(
            UserSerializer(user).data,
            status=status.HTTP_201_CREATED,
            headers=headers,
        )


class UserEmailVerification(APIView):
    """
    Verify a user's email address
    """

    @extend_schema(
        operation_id="VerifyEmail",
        tags=["users"],
        parameters=[
            OpenApiParameter(
                "email",
                description="E-mail to verify",
                type=OpenApiTypes.EMAIL,
                required=True,
            ),
            OpenApiParameter(
                "token",
                description="Verification token",
                required=True,
            )
        ],
        responses={302: serializers.Serializer}
    )
    def get(self, *args, **kwargs):
        if not all(arg in self.request.GET for arg in ("email", "token")):
            raise ValidationError

        try:
            user = User.objects.get_by_natural_key(self.request.GET["email"])
        except User.DoesNotExist:
            raise AuthenticationFailed

        if not default_token_generator.check_token(user, self.request.GET["token"]):
            raise AuthenticationFailed

        user.verified_email = True
        user.save()
        UserScope.objects.bulk_create([
            UserScope(user=user, scope=Scope.CreateIIIFImage),
            UserScope(user=user, scope=Scope.UploadS3Image),
        ])

        if user.has_usable_password():
            login(self.request, user)
            # A user with a verified email is granted a private trial corpus with admin rights
            if not Corpus.objects.admin(user=user).exists():
                corpus = Corpus.objects.create(
                    name="My Project",
                    description=f"Project for {user.email}"
                )
                corpus.memberships.create(
                    user=user,
                    level=Role.Admin.value
                )
                # Add default types
                corpus.create_default_types()

            # Redirtect to home page
            url = "/"
        else:
            url = reverse("password_reset_confirm", kwargs={
                "uidb64": urlsafe_base64_encode(str(user.id).encode()),
                "token": self.request.GET["token"],
            })

        # If a default group into which to add newly registered and verified users has been
        # defined, then add the user to that group
        if settings.SIGNUP_DEFAULT_GROUP:
            Group.objects.get(pk=settings.SIGNUP_DEFAULT_GROUP).memberships.create(user=user, level=Role.Guest.value)

        return Response(
            status=status.HTTP_302_FOUND,
            headers={"Location": url},
        )


@extend_schema_view(post=extend_schema(operation_id="ResetPassword", tags=["users"]))
class PasswordReset(CreateAPIView):
    """
    Start a password reset flow
    """
    serializer_class = PasswordResetSerializer

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        form = PasswordResetForm(serializer.validated_data)

        if form.is_valid():
            form.save(
                request=self.request,
                use_https=self.request.is_secure(),
                token_generator=default_token_generator,
                email_template_name="registration/password_reset_email.html",
                subject_template_name="registration/password_reset_subject.txt",
            )
            headers = self.get_success_headers(serializer.data)
            return Response(
                status=status.HTTP_201_CREATED,
                headers=headers,
            )
        else:
            return Response(
                form.errors.get_json_data(),
                status=status.HTTP_400_BAD_REQUEST,
                headers=self.default_response_headers
            )


@extend_schema_view(post=extend_schema(operation_id="PasswordResetConfirm", tags=["users"]))
class PasswordResetConfirm(CreateAPIView):
    """
    Confirm a password reset using data from the confirmation e-mail
    """
    serializer_class = PasswordResetConfirmSerializer


@extend_schema(tags=["users"])
class GroupsList(ListCreateAPIView):
    """
    List existing groups either if they are public or the request user is a member.
    A POST request allow to create a new group.
    """
    serializer_class = GroupSerializer
    permission_classes = (IsVerified, )

    def get_queryset(self):
        # List public groups and ones to which user belongs
        return Group.objects \
            .annotate(members_count=Count("users")) \
            .filter(Q(public=True) | Q(users__in=[self.request.user])) \
            .order_by("name", "id")


@extend_schema(tags=["jobs"])
class JobList(ListAPIView):
    """
    List asynchronous jobs linked to the current user.
    """
    permission_classes = (IsVerified, )
    serializer_class = JobSerializer
    pagination_class = None

    def get_queryset(self):
        if not self.request:
            return []
        return self.request.user.get_rq_jobs()


@extend_schema(
    tags=["jobs"],
    parameters=[
        OpenApiParameter(
            "id",
            location=OpenApiParameter.PATH,
            type=OpenApiTypes.UUID,
        )
    ]
)
class JobRetrieve(RetrieveDestroyAPIView):
    """
    Retrieve a single job by ID.
    """
    permission_classes = (IsVerified, )
    serializer_class = JobSerializer

    def get_object(self):
        for queue_name in QUEUES.keys():
            job = get_queue(queue_name).fetch_job(str(self.kwargs["pk"]))
            if not job:
                continue
            if job.user_id != self.request.user.id:
                raise NotFound
            return job
        raise NotFound

    def perform_destroy(self, instance):
        if instance.get_status(refresh=False) == JobStatus.STARTED:
            raise ValidationError(["Cannot delete a running job."])
        instance.delete()


@extend_schema(tags=["users"])
class GroupsCreate(CreateAPIView):
    """
    Create a new group. The request user will be added as a member of this group.
    """
    serializer_class = GroupSerializer
    permission_classes = (IsVerified, )
    # For OpenAPI type discovery
    queryset = Group.objects.none()


@extend_schema(tags=["users"])
@extend_schema_view(
    patch=extend_schema(
        description="Partially update details of a group. "
                    "Requires to have `admin` privileges on this group."
    ),
    put=extend_schema(
        description="Update details of a group. "
                    "Requires to have `admin` privileges on this group."
    ),
    delete=extend_schema(
        description="Delete a group. "
                    "Requires to have `admin` privileges on this group."
    )
)
class GroupDetails(RetrieveUpdateDestroyAPIView):
    """
    Retrieve details about a specific group
    """
    serializer_class = GroupSerializer
    permission_classes = (IsVerified, )
    # For OpenAPI type discovery
    queryset = Group.objects.none()

    def get_membership(self, group):
        try:
            return group.memberships.get(user_id=self.request.user.id)
        except Right.DoesNotExist:
            raise PermissionDenied(detail="Only members of a group can retrieve their details")

    def get_object(self):
        if not hasattr(self, "_group"):
            self._group = super().get_object()

            if self.request.user.is_admin:
                self._group.level = Role.Admin.value
            else:
                # Retrieve the membership to know the privilege level for this user
                if not hasattr(self, "_member"):
                    self._member = self.get_membership(self._group)
                # Add request member level to the group
                self._group.level = self._member.level

        return self._group

    def check_object_permissions(self, request, obj):
        if request.user.is_admin:
            return
        if not hasattr(self, "_member"):
            self._member = self.get_membership(obj)
        # Check the user has the right to delete or update a group
        super().check_object_permissions(request, obj)
        if request.method not in SAFE_METHODS and self._member.level < Role.Admin.value:
            raise PermissionDenied(detail="Only members with an admin privilege may update or delete this group.")

    def get_queryset(self):
        # Superusers have access to all groups
        if self.request.user.is_authenticated and self.request.user.is_admin:
            return Group.objects.all() \
                .annotate(members_count=Count("memberships"))
        # Filter groups that are public or for which the user is a member
        return Group.objects \
            .annotate(members_count=Count("memberships")) \
            .filter(
                Q(public=True)
                | Q(memberships__user=self.request.user)
            )


@extend_schema(tags=["users"])
class UserMemberships(ListAPIView):
    """
    List groups to which the request user belongs with its privileges on each group
    """
    serializer_class = MemberGroupSerializer
    permission_classes = (IsVerified, )

    def get_queryset(self):

        def _build_fake_right(group):
            # Serialize a fake memberships for admin user with no needs for ID and level
            admin_membership = Right(
                content_object=group,
                user=self.request.user,
                level=None,
            )
            admin_membership.id = None
            admin_membership.members_count = group.members_count
            return admin_membership

        if self.request.user.is_authenticated and self.request.user.is_admin:
            # Allow super admins to access all groups as if they had an admin access level
            groups = Group.objects.all() \
                .annotate(members_count=Count("memberships")) \
                .order_by("name", "id")
            return [_build_fake_right(group) for group in groups]
        # Annotate rights with the group member count as Prefetch is not available with the Generic FK
        return self.request.user.rights \
            .prefetch_related("content_object") \
            .filter(content_type=ContentType.objects.get_for_model(Group)) \
            .annotate(members_count=Count("group_target__memberships")) \
            .order_by("group_target__name", "group_target__id")


@extend_schema(tags=["users"])
@extend_schema_view(
    patch=extend_schema(
        description="Partially update a generic membership."
    ),
    put=extend_schema(
        description="Update a generic membership."
    ),
    delete=extend_schema(
        description="Delete a generic membership."
    )
)
class MembershipDetails(ACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Retrieve details of a generic membership
    """
    serializer_class = MembershipSerializer
    permission_classes = (IsVerified, )
    # Perform access checks with the permissions
    queryset = Right.objects.select_related("content_type").all()

    def check_object_permissions(self, request, membership):
        super().check_object_permissions(request, membership)

        # Retrieve access level
        access_level = None
        if isinstance(membership.content_object, Worker):
            # Use WorkerACLMixin method as the worker access level may be inherited from its repository
            access_level = WorkerACLMixin.get_max_level(self, membership.content_object)
        else:
            access_level = get_max_level(self.request.user, membership.content_object)

        # Check the access level of the request user on the content object
        if not access_level or access_level < Role.Guest.value:
            raise NotFound

        #  Allow any guest to retrieve information
        if self.request.method in SAFE_METHODS:
            return

        # Allow a user to remove its own membership
        if self.request.method == "DELETE" and membership.user_id == self.request.user.id:
            return

        # Only allow admins to edit the access level of a membership
        if access_level < Role.Admin.value:
            raise PermissionDenied(
                detail="Only admins of the target membership group can perform this action."
            )

    def perform_destroy(self, instance):
        # At least one admin member must exist except for workers which depends on their repositories
        if (
            instance.content_type != ContentType.objects.get_for_model(Worker)
            and instance.level >= Role.Admin.value
            and instance.content_object.memberships.using("default").filter(level__gte=Role.Admin.value).count() < 2
        ):
            raise ValidationError({"detail": ["Removing all memberships with an admin privilege is not possible."]})
        return super().perform_destroy(instance)


MEMBERSHIPS_TYPE_FILTERS = ["user", "group"]


@extend_schema(
    parameters=[
        OpenApiParameter(
            content.name,
            type=OpenApiTypes.UUID,
            description=f"List user members for a {content}",
        )
        for content in RightContent
    ] + [
        OpenApiParameter(
            "type",
            description="Filter memberships by owner type",
            enum=MEMBERSHIPS_TYPE_FILTERS,
        )
    ],
    tags=["users"]
)
class GenericMembershipsList(WorkerACLMixin, ListAPIView):
    """
    List memberships for a specific content with their privileges.
    Target of the right must be defined as a query parameter.
    """
    permission_classes = (IsVerified, )
    serializer_class = GenericMembershipSerializer

    def get_content_object(self):
        """
        Retrieve a generic content based on query parameters.
        This method does not perform any permission check.
        """
        allowed_contents = [c.name for c in RightContent]
        content_keys = [key for key in self.request.query_params if key in allowed_contents]
        # Assert exactly one of the content object parameter has been set
        if len(content_keys) != 1:
            raise ValidationError({
                "__all__": [
                    "Exactly one of those query parameters must be defined: "
                    f"{', '.join(allowed_contents)}."
                ]
            })

        key, = content_keys
        content_model = RightContent[key].value
        content_id = self.request.query_params[key]
        try:
            content_uuid = UUID(content_id)
        except (AttributeError, ValueError):
            raise ValidationError({key: [f"'{content_id}' is not a valid UUID."]})

        # Use default database in case the content object has just been created
        return get_object_or_404(content_model.objects.using("default"), id=content_uuid)

    def get_queryset(self):
        content_object = self.get_content_object()

        # Determine the user right on the generic content
        has_access = None
        if isinstance(content_object, Worker):
            # Workers rights may be determined by their repository
            has_access = self.has_worker_access(content_object, Role.Guest.value)
        else:
            access_params = {}
            if isinstance(content_object, Group):
                # A user may not be able to list group members if he is not a member
                access_params["skip_public"] = True
            has_access = self.has_access(content_object, Role.Guest.value, **access_params)

        # Ensure the user has a guest access to the content
        if not has_access:
            raise PermissionDenied(
                detail="You do not have the required access level to list members for this content."
            )

        # Avoid a stale read when adding/deletig a member
        qs = content_object.memberships \
            .using("default") \
            .prefetch_related(
                "user",
                Prefetch("group", queryset=Group.objects.annotate(members_count=Count("memberships")))
            ) \
            .order_by("user__display_name", "group__name", "id")

        # Handle a simple owner type filter
        type_filter = self.request.query_params.get("type")
        if type_filter and type_filter in MEMBERSHIPS_TYPE_FILTERS:
            type_filter = {f"{type_filter}__isnull": False}
            qs = qs.filter(**type_filter)

        return qs


@extend_schema_view(
    post=extend_schema(
        operation_id="CreateMembership",
        tags=["users"],
    )
)
class GenericMembershipCreate(CreateAPIView):
    """
    Create a new generic membership.
    """
    permission_classes = (IsVerified, )
    serializer_class = MembershipCreateSerializer
